fun id x = x
fun fst (x,y) = x;
let 
    val a = (id,id)     val a = (a,a) 
    val a = (a,a)     val a = (a,a) 
    val a = (a,a)     val a = (a,a) 
    val a = (a,a)     val a = (a,a) 
    val a = (a,a)     val a = (a,a) 
    val a = (a,a)     val a = (a,a) 
    val a = (a,a)     val a = (a,a) 
    val a = (a,a)     val a = (a,a) 
    val a = (a,a)     val a = (a,a) 
in 
fst (fst(
fst (fst(
fst (fst(
fst (fst(
fst (fst(
fst (fst(
fst (fst(
fst (fst(
fst (fst( 
a
) )
) )
) )
) )
) )
) )
) )
) )
) )
1
end
;
