let 
    val a = fn x => x
    val a = fn x => a a x
    val a = fn x => a a x
    val a = fn x => a a x
    val a = fn x => a a x
    val a = fn x => a a x
    val a = fn x => a a x
    val a = fn x => a a x
    val a = fn x => a a x
    val a = fn x => a a x
    val a = fn x => a a x
    val a = fn x => a a x
    val a = fn x => a a x
    val a = fn x => a a x
    val a = fn x => a a x
    val a = fn x => a a x
    val a = fn x => a a x
    val a = fn x => a a x
    val a = fn x => a a x
    val a = fn x => a a x
    val a = fn x => a a x
    val a = fn x => a a x
    val a = fn x => a a x
    val a = fn x => a a x
    val a = fn x => a a x
in 
a 1
end
;
